from builtins import str

from django import forms
from django.forms.widgets import TextInput
from django.core.exceptions import ValidationError

from .version import Version
from .constants import DEFAULT_NUMBER_BITS
from .utils import convert_version_int_to_string


class VersionField(forms.IntegerField):
    widget = TextInput

    def __init__(self, number_bits=DEFAULT_NUMBER_BITS, **kwargs):
        self.number_bits = number_bits
        return super(VersionField, self).__init__(**kwargs)

    def to_python(self, value):
        """
        Verifies that value can be converted to a Version object
        """
        if not value:
            return None

        if isinstance(value, str):
            try:
                return int(Version(value, self.number_bits))
            except ValueError:
                max_value = '.'.join([str(2 ** e - 1) for e in self.number_bits])
                raise ValidationError("Max version is {0}".format(max_value))

        return Version(convert_version_int_to_string(value, self.number_bits), self.number_bits)

    def widget_attrs(self, widget):
        attrs = super(VersionField, self).widget_attrs(widget)
        attrs['pattern'] = '^' + (r'(\d+\.)?' * (len(self.number_bits) - 1)) + r'(\*|\d+)$'
        return attrs



def convert_version_string_to_int(string, number_bits):
    """
    Take in a verison string e.g. '3.0.1'
    Store it as a converted int: 3*(2**number_bits[0])+0*(2**number_bits[1])+1*(2**number_bits[2])

    >>> convert_version_string_to_int('3.0.1',[8,8,16])
    50331649
    """
    numbers = [int(number_string) for number_string in string.split(".")]

    if len(numbers) > len(number_bits):
        raise NotImplementedError(
            "Versions with more than {0} decimal places are not supported".format(
                len(number_bits) - 1))

    # add 0s for missing numbers
    numbers.extend([0] * (len(number_bits) - len(numbers)))

    #convert to single int and return
    number = 0
    total_bits = 0
    for num, bits in reversed(list(zip(numbers, number_bits))):
        max_num = (bits + 1) - 1
        if num >= 1 << max_num:
            raise ValueError(
                "Number {0} cannot be stored with only {1} bits. Max is {2}".format(num, bits,
                                                                                    max_num))
        number += num << total_bits
        total_bits += bits

    return number


def convert_version_int_to_string(number, number_bits):
    """
    Take in a verison string e.g. '3.0.1'
    Store it as a converted int: 3*(2**number_bits[0])+0*(2**number_bits[1])+1*(2**number_bits[2])

    >>> convert_version_int_to_string(50331649,[8,8,16])
    '3.0.1'
    """
    number_strings = []
    total_bits = sum(number_bits)
    for bits in number_bits:
        shift_amount = (total_bits - bits)
        number_segment = number >> shift_amount
        number_strings.append(str(number_segment))
        total_bits = total_bits - bits
        number = number - (number_segment << shift_amount)

    return ".".join(number_strings)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup, find_packages
except ImportError:
    import ez_setup

    ez_setup.use_setuptools()
    from setuptools import setup, find_packages

import os

setup(
    name="django-versionfield3",
    version="0.1.3",
    url="https://github.com/Crystalnix/django-versionfield3",
    license="BSD",
    description="A DB Independent Custom Django Field for storing Version numbers for fast indexing",
    author="Egor Yurtaev",
    author_email="yurtaev.egor+versionfield@gmail.com",
    packages=["versionfield"],
    include_package_data=True,
    install_requires=[
        "future>=0.16.0",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Internet :: WWW/HTTP",
    ]
)



# coding: utf8
from builtins import str
from past.builtins import basestring
from future.utils import python_2_unicode_compatible

from .utils import convert_version_string_to_int, convert_version_int_to_string


class Version(object):
    def __init__(self, string, number_bits):
        """
        Take in a verison string e.g. '3.0.1'
        Store it as a converted int
        """
        self.number_bits = number_bits
        self.internal_integer = convert_version_string_to_int(string, number_bits)

    @python_2_unicode_compatible
    def __str__(self):
        return str(convert_version_int_to_string(self.internal_integer, self.number_bits))

    def __repr__(self):
        return self.__str__()

    def __int__(self):
        return self.internal_integer

    def __eq__(self, other):
        if not other:
            return False  # we are obviously a valid Version, but 'other' isn't
        if isinstance(other, basestring):
            return self == Version(other, self.number_bits)
        else:
            return int(self) == int(other)



from builtins import str
from past.builtins import basestring
from future.utils import python_2_unicode_compatible

import django
from django.db import models
from django.core.exceptions import ValidationError

from .constants import DEFAULT_NUMBER_BITS
from .version import Version
from .utils import convert_version_string_to_int, convert_version_int_to_string
from . import forms

if django.VERSION[:2] <= (1, 7):
    from django.utils.six import with_metaclass


    class BaseField(with_metaclass(models.SubfieldBase, models.BigIntegerField)):
        def to_python(self, value):
            if isinstance(value, Version):
                return int(value)

            if isinstance(value, basestring):
                return Version(value, self.number_bits)

            if value is None:
                return None

            return Version(convert_version_int_to_string(value, self.number_bits), self.number_bits)
else:
    class BaseField(models.BigIntegerField):
        def from_db_value(self, value, *args, **kwargs):
            if isinstance(value, Version):
                return int(value)

            if isinstance(value, basestring):
                return Version(value, self.number_bits)

            if value is None:
                return None

            return Version(convert_version_int_to_string(value, self.number_bits), self.number_bits)


class VersionField(BaseField):
    """
    A Field where version numbers are input/output as strings (e.g. 3.0.1)
    but stored in the db as converted integers for fast indexing
    """
    description = "A version number (e.g. 3.0.1)"

    def __init__(self, number_bits=DEFAULT_NUMBER_BITS, *args, **kwargs):
        self.number_bits = number_bits
        super(VersionField, self).__init__(*args, **kwargs)

    def get_prep_value(self, value):
        if isinstance(value, basestring):
            try:
                return int(Version(value, self.number_bits))
            except ValueError:
                max_value = '.'.join([str(2 ** e - 1) for e in self.number_bits])
                raise ValidationError("Max version is {0}".format(max_value))

        if value is None:
            return None

        return int(value)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': forms.VersionField,
            'number_bits': self.number_bits
        }
        defaults.update(kwargs)
        return super(VersionField, self).formfield(**defaults)

    @python_2_unicode_compatible
    def __str__(self, value):
        return str(value)


try:
    from south.modelsinspector import add_introspection_rules

    rules = [
        (
            (VersionField,),
            [],
            {
                "number_bits": ["number_bits", {"default": DEFAULT_NUMBER_BITS}],
            },
        )
    ]
    add_introspection_rules(rules, ["^versionfield"])
except ImportError:
    # looks like we aren't using south
    pass
